/* -*- mode: c++; c-basic-offset: 4; -*- */

/**  llmonit generates statistics frames for one or more low-latency partitions.
  *  Two time series are generated for each registered partititon - a dropout
  *  count and a latency time.
  */
#include "gmutex.hh"
#include "thread_base.hh"
#include "FixedLenTS.hh"
#include "FrWriter.hh"
#include "frame_name.hh"
#include <string>
class LSMP_CON;

/**  ll_collect is a thread that monitors a single partition.
 */
class ll_collect : public thread::thread_base {
public:
    /**  The %_tswin class contains a time series window for the collector.
      *  \brief %Time series window class.  
      */
    class _tswin {
    public:
	/**  Construct a fixed length time series window. In addition to the
	  *  data inside the window, each instance contains the window 
	  *  parameters (sample period and length) and a read/write mutex
	  *  to use in synchronizing the data access.
	  *  \brief %Time series window class.
	  *  \param delta %Time series sample interval.
	  *  \param max_length Maximum lengh of window (in seconds). 
	  */
	_tswin(Interval delta=4, Interval max_length=28800);

	/** Add a time sample to the time series.
	 */
	void add(const Time& t, float v);
	const std::string& channel(void) const;
	void clear(void);
	Interval dT(void) const;
	TSeries extract(const Time& t, Interval dt) const;
	void set_channel(const std::string& chan);
    private:
	Interval    _dt;
	FixedLenTS  _window;
	std::string _channel;
	mutable thread::readwritelock _mutex;
    };

public:
    /**  Monitor constructor take the main function arguments.
      */
    ll_collect(const std::string& node, const std::string& name);
    ~ll_collect(void);
    void connect(void);
    void disconnect(void);
    const _tswin& ref_late(void) const;
    const _tswin& ref_long(void) const;
    const _tswin& ref_recv(void) const;
    void set_partition(const std::string& name);
    const std::string& partition(void) const;
    void* thread_entry(void);

private:
    std::string _part;         ///< partition name
    std::string _node;         ///< node name
    LSMP_CON*   _cons;         ///< lsmp consumer pointer
    _tswin      _recv;         ///< Dropout record
    _tswin      _late;         ///< Latency record
    _tswin      _length;       ///< Length record
};

//======================================  Collector inline functions
inline const ll_collect::_tswin& 
ll_collect::ref_late(void) const {
    return _late;
}

inline const ll_collect::_tswin& 
ll_collect::ref_long(void) const {
    return _length;
}

inline const ll_collect::_tswin& 
ll_collect::ref_recv(void) const {
    return _recv;
}

inline const std::string&
ll_collect::_tswin::channel(void) const {
    return _channel;
}

inline Interval
ll_collect::_tswin::dT(void) const {
    return _dt;
}

/** Collect_list contains a list of running collectors.
 */
typedef std::vector<ll_collect*>     collect_list;
typedef collect_list::iterator       collect_iter;
typedef collect_list::const_iterator const_collect_iter;

/**  ll_writer is a thread that writes data from all collectors into a single
  *  frame.
  */
class ll_writer : public thread::thread_base {
public:
    /**  Construct a frame writer. 
      */
    ll_writer(Interval dt, const collect_list& colist);
    ~ll_writer(void);
    frame_name& pattern(void);
    bool is_open(void) const;
    void open(const std::string& name, int runId);
    void set_delay(Interval delay);
    void set_frame_path(const std::string& path);
    void* thread_entry(void);
    void write(const Time& t);
private:
    void addChannel(const Time& t, const ll_collect::_tswin& cl);
private:
    Interval            _dt;
    Interval            _delay;
    const collect_list& _collect;
    frame_name          _pattern;
    FrWriter            _writer;
    Time                _lastGPS;
};

//======================================  Inline llwrite
inline frame_name&
ll_writer::pattern(void) {
    return _pattern;
}


/** The llmonit class it the monitor class for this monitor.
 */
class llmonit {
public:
    /**  Monitor constructor take the main function arguments.
     */
    llmonit(int argc, const char* argv[]);
    ~llmonit(void);
    void configure(void);
    void run(void);
    void shut(void);
private:
    std::string  _config;
    std::string  _outdir;
    std::string  _name;
    int          _RunID;
    Interval     _dt;
    ll_writer*   _writer;
    collect_list _collectors;
};

//======================================  llthread accessors.
inline const std::string& 
ll_collect::partition(void) const {
  return _part;
}
